#ifndef AVLTREE
#define AVLTREE

template<class T>
class AVLTree
{
    protected:

        struct tree_node
        {
            T item;
            tree_node* parent,
                     * left,
                     * right;
            bool isHeader;
            char balanceFactor;
        }; // tree_node

        typedef tree_node* Link;

        Link header;
        long node_count;

    public:

        class Iterator;
        friend class Iterator;


        // Postcondition: this AVLTree is empty.
        AVLTree ()
        {
            header = new tree_node;
            header -> parent = NULL;
            header -> left = header;
            header -> right = header;
            header -> isHeader = true;
            node_count = 0;
        } //  default constructor


        // Postcondition: the number of items in this AVLTree
        //                has been returned.
        unsigned size() const
        {
              return node_count;
        } // size

        // Postcondition: true has been returned if this tree
        //                really is an AVLTree.  Otherwise, false
        //                has been returned.
        bool isAVLTree() { return checkAVLTree( header -> parent ); }


        // Postcondition: the AVL property has been restored, if
        //                necessary, by rotations and balance-factor
        //                adjustments between the inserted tree_node and
        //                its nearest ancestor with a balanceFactor of
        //                'L' or 'R'.
        void fixAfterInsertion (Link ancestor, Link inserted)
        {
            Link root = header -> parent;

            T item = inserted -> item;

            if (ancestor == NULL)
            {
                if (item < root -> item)
                    root -> balanceFactor = 'L';
                else
                    root -> balanceFactor = 'R';
                adjustPath (root, inserted);
            } // all ancestor's balance factors are '='
            else if ((ancestor -> balanceFactor == 'L' &&
                      item >= ancestor -> item) ||
                     (ancestor -> balanceFactor == 'R' &&
                      item < ancestor -> item))
            {
                 ancestor -> balanceFactor = '=';
                 adjustPath (ancestor, inserted);
            } // insertion in opposite subtree of ancestor's balance factor
            else if (ancestor -> balanceFactor == 'R' &&
                     item >= ancestor -> right -> item)
            {
                 ancestor -> balanceFactor = '=';
                 rotate_left (ancestor);
                 adjustPath (ancestor -> parent, inserted);
            } // insertion in right subtree of ancestor's right subtree
            else if (ancestor -> balanceFactor == 'L' &&
                     item < ancestor -> left -> item)
            {
                ancestor -> balanceFactor = '=';
                rotate_right (ancestor);
                adjustPath (ancestor -> parent, inserted);
            } // insertion in left subtree of ancestor's left subtree
            else if (ancestor -> balanceFactor == 'L' &&
                     item >= ancestor -> left -> item)
            {
                rotate_left (ancestor -> left);
                rotate_right (ancestor);
                adjustLeftRight (ancestor, inserted);
            } // insertion in right subtree of ancestor's left subtreee
            else
            {
                rotate_right (ancestor -> right);
                rotate_left (ancestor);
                adjustRightLeft (ancestor, inserted);
            } // insertion in left subtree of ancestor's right subtree
        } // method fixAfterInsertion


        // Postcondition: the balanceFactors of all entries between the
        //                inserted node (exclusive) and the to node (exclusive)
        //                have been adjusted, if necessary.
        void adjustPath (Link to, Link inserted) {

            T item = inserted -> item;

            Link temp = inserted -> parent;
            while (temp != to)
            {
                if (item < temp -> item)
                    temp -> balanceFactor = 'L';
                else
                    temp -> balanceFactor = 'R';
                temp = temp -> parent;
            } // while
        } // method adjustPath


        // Postcondition: all of the balance factors in the path from inserted
        //                (exclusive) to ancestor's sibling (inclusive) have
        //                been adjusted after a left-right rotation.
        void adjustLeftRight (Link ancestor, Link inserted) {

            T item = inserted -> item;

            if (ancestor -> parent == inserted) // case 3a
                ancestor -> balanceFactor = '=';
            else if (item < ancestor -> parent -> item)
            {
                ancestor -> balanceFactor = 'R';
                adjustPath (ancestor -> parent -> left, inserted);
            } // case 3b: item >= ancestor's parent's item
            else
            {
                ancestor -> balanceFactor = '=';
                ancestor -> parent -> left -> balanceFactor = 'L';
                adjustPath (ancestor, inserted);
            } // case 3c: item < ancestor's parent's item
        } // method adjustLeftRight


        // Postcondition: all of the balance factors in the path from inserted
        //                (exclusive) to ancestor's sibling (inclusive) have
        //                been adjusted after a right-left rotation.
        void adjustRightLeft (Link ancestor, Link inserted)
        {
            T item = inserted -> item;

            if (ancestor -> parent == inserted) // case 4a
                ancestor -> balanceFactor = '=';
            else if (item >= ancestor -> parent -> item)
            {
                ancestor -> balanceFactor = 'L';
                adjustPath (ancestor -> parent -> right, inserted);
            } // case 4b: item >= ancestor's parent's item
            else
            {
                ancestor -> balanceFactor = '=';
                ancestor -> parent -> right -> balanceFactor = 'R';
                adjustPath (ancestor, inserted);
            } // case 4b: item < ancestor's parent's item
        } // method adjustRightLeft


       void rotate_left(Link x)
       {
           Link y = x -> right;
           x -> right = y -> left;
           if (y -> left != NULL)
               y -> left -> parent = x;
           y -> parent = x -> parent;
           if (x == header -> parent)
               header -> parent = y;
           else if (x == x -> parent -> left)
               x -> parent -> left = y;
           else
               x -> parent -> right = y;
           y -> left = x;
          x -> parent = y;
       }

       void rotate_right(Link x) {
           Link y = x -> left;
           x -> left = y -> right;
           if (y -> right != NULL)
               y -> right -> parent = x;
           y -> parent = x -> parent;
           if (x == header -> parent)
               header -> parent = y;
           else if (x == x -> parent -> right)
               x -> parent -> right = y;
           else
               x -> parent -> left = y;
           y -> right = x;
           x -> parent = y;
       }


        // Postcondition: item has been inserted into the tree. The worstTime
        //                (n) is O (n) and averageTime (n) is O (log n).
        Iterator insert (const T& item)
        {
            if (header -> parent == NULL)
            {
                insertItem (item, header, header -> parent);
                header -> left = header -> parent;
                header -> right = header -> parent;
                return header -> parent;
             } // inserting at tree's root
             else
             {
                 Link parent = header,
                               child = header -> parent,
                               ancestor = NULL;

                 while (child != NULL)
                 {
                     parent = child;
                     if (child -> balanceFactor != '=')
                         ancestor = child;
                     if (item < child -> item)
                         child = child -> left;
                     else
                         child = child -> right;
                 } // while
                 if (item < parent -> item)
                 {
                     insertItem (item, parent, parent -> left);
                     fixAfterInsertion (ancestor, parent -> left);
                     if (header -> left == parent)
                          header -> left = parent -> left;  // child is leftmost
                     return parent -> left;
                 } // insert at left of parent
                 else
                 {
                     insertItem (item, parent, parent -> right);
                     fixAfterInsertion (ancestor, parent -> right);
                     if (header -> right == parent)
                         header -> right = parent -> right; // child is rightmost
                     return parent -> right;
                 } // insert at right of parent
             } // tree not empty
         } // insert


         // Postcondition: the height of this AVLTree has been
         //                returned.  The worstTime (n) is O (n).
         int height() { return height( header -> parent ); }


         class Iterator
         {
             friend class AVLTree<T>;

             protected:

                 Link link;
                 Iterator (Link new_link) : link (new_link) {}

             public:

                 Iterator () {}
                 Iterator& operator++ ()
                 {
                      Link tempLink;

                      if ((link -> right) != NULL)
                      {
                          link = link -> right;
                          while ((link -> left) != NULL)
                              link = link -> left;
                      } // node has right child
                      else
                      {
                          tempLink = link -> parent;
                          while (link == tempLink -> right)
                          {
                              link = tempLink;
                              tempLink = tempLink -> parent;
                          } // moving up and leftward as far as possible
                          if ((link -> right) != tempLink)
                              link = tempLink;
                      } // node has no right child
                      return *this;
                 } // prefix ++

                 Iterator operator++ (int)
                 {
                     Iterator temp = *this;
                     ++(*this);
                     return temp;
                 } // postfix ++

                 Iterator& operator-- ()
                 {
                     if (link -> isHeader)
                        link = link -> right;   // Return rightmost
                     else if (link -> left != NULL)
                     {
                         Link y = link -> left;
                         while (y -> right != NULL)
                             y = y -> right;
                         link = y;
                     } // link -> left != NULL
                     else
                     {
                         Link y = link -> parent;
                         while (link == y -> left)
                         {
                             link = y;
                             y = y -> parent;
                         } // while moving up and rightward
                         link = y;
                     } // link -> left == NULL
                     return *this;
             } // pre-decrement operator

            T& operator* () const { return link -> item; }
            bool operator== (const Iterator &otherIterator)const
            {
                return link == otherIterator.link;
            } //operator ==
            bool operator!= (const Iterator &otherIterator)const
            {
                return link != otherIterator.link;
            } // operator !=

      }; // Iterator


      // Postcondition: an iterator positioned at the beginning of
      //                this AVLTree has been returned.
      Iterator begin () { return header -> left; }


      // Postcondition: an iterator that can be used in comparisons
      //                to terminate an iteration through this
      //                AVLTree has been returned.
      Iterator end () { return header; }


      // Postcondition: If there is an item in the tree
      //                that is == Item, the iterator returned points to that
      //                item.  Otherwise, the iterator returned points past the
      //                end of the tree.  The worstTime (n) is O (n) and
      //                averageTime (n) is O (log n).
      Iterator find (const T& item)
      {
          Link parent = header;
          Link child = header -> parent;
          while (child != NULL)
          {
              if (item <= child -> item)
              {
                  parent = child;
                  child = child -> left;
              } // item <= child -> item
              else
                  child = child -> right;
          } // while

          if (parent == header || item != parent -> item)
              return end();
          else
              return Iterator (parent);
      } // find

      
      // Precondition: itr is positioned at an item in this AVLTree.
      // Postcondition: the item that itr was positioned at has been
      //                removed from this AVLTree.  The worstTime is
      //                O (n) and averageTime (n) is O (log n).
      void erase (Iterator itr)
      {
          if (itr.link -> parent -> parent == itr.link)
              deleteLink (itr.link -> parent -> parent);
          else if (itr.link -> parent -> left == itr.link)
              deleteLink (itr.link -> parent -> left);
          else
              deleteLink (itr.link -> parent -> right);
       } // erase


    protected:

        // Postcondition: A copy of item has been inserted in the tree at the
        //                child of the given parent.
        void insertItem (const T& item, Link parent, Link& child)
        {
            child = new tree_node;
            child -> item = item;
            child -> parent = parent;
            child -> left = NULL;
            child -> right = NULL;
            child -> isHeader = false;
            child -> balanceFactor = '=';
            node_count++;
        } // insertItem


        // Postcondition: The subtree pointed to by link has been deleted.
        void prune (Link& link)
        {
            Link linkCopy = link,
                 newLink;

            node_count--;
            if ((link -> left == NULL) && (link -> right == NULL))
            {
                if (link == header -> left)
                    header -> left = link -> parent; // new leftmost
                if (link == header -> right)
                    header -> right = link -> parent; // new rightmost
                link = NULL;
            } // link's node is a leaf
            else if (link -> left == NULL)
            {
                link = link -> right;
                link -> parent = linkCopy -> parent;
                if (linkCopy == header -> left)
                {
                    newLink = link;
                    while ((newLink -> left) != NULL)
                        newLink = newLink -> left;
                    header -> left = newLink;  // new leftmost
                } // re-calculate leftmost
            } // link -> left nonempty
            else
            {
                 link = link -> left;
                 link -> parent = linkCopy -> parent;
                 if (linkCopy == header -> right)
                 {
                      newLink = link;
                      while ((newLink -> right) != NULL)
                          newLink = newLink -> right;
                      header -> right = newLink; // new rightmost
                 } // re-calculate rightmost
            } // link -> right nonempty
            delete linkCopy;
       } // prune


       // Postcondition: successor contains the item that, before it was
       //                deleted during this call, was the successor of
       //                the tree's link item.
       void deleteSuccessor (T& successor, Link& link)
       {
            if (link -> left == NULL)
            {
                 successor = link -> item;
                 prune (link);
            } // empty left subtree
            else
            {
                 Link temp = link;
                 while (temp -> left != NULL)
                     temp = temp -> left;
                 successor = temp -> item;
                 prune (temp -> parent -> left);
            } // nonempty left subtree
       } // deleteSuccessor


       // Postcondition: The item that had been the link item has been
       //                deleted.
       void deleteLink (Link& link)
       {
            T successor;

            if (link -> left == NULL || link -> right == NULL)
                prune (link);
            else
            {
                deleteSuccessor (successor, link -> right);
                link -> item = successor;
            } // neither subtree empty
       } // deleteLink


       // Postcondition: the height of the subtree starting at root
       //                has been returned.
       int height (const Link& root)
       {
           int leftHeight,
               rightHeight;

           if (root == NULL)
               return -1;
           else
           {
               leftHeight = height (root -> left);
               rightHeight = height (root -> right);
               if (leftHeight > rightHeight)
                   return 1 + leftHeight;
               return 1 + rightHeight;
           } // else
       } // height


    // Postcondition: The tree has been checked to make sure that the
    //                heights of the left and right subtrees (if they
    //                exist) are within 1 of each other, and that those
    //                subtrees are AVL trees.
    bool checkAVLTree (Link root)
    {
	if (root == NULL)
            return true;
	else
            if ((abs (height (root -> left) - height (root -> right)) <= 1)
                 && (checkAVLTree (root -> left))
                 && (checkAVLTree (root -> right)))
            return true;
	return false;
    } // checkAVLTree
}; // AVLTree


#endif
